/* $Id: tif_unix.c,v 1.21 2007/07/18 13:46:41 dron Exp $ */

/*
 * Copyright (c) 1988-1997 Sam Leffler
 * Copyright (c) 1991-1997 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Sam Leffler and Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Sam Leffler and Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SAM LEFFLER OR SILICON GRAPHICS BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THIS SOFTWARE.
 */

/*
 * TIFF Library UNIX-specific Routines. These are should also work with the
 * Windows Common RunTime Library.

Modified by Alexey Chupahin
Rostov-on-Don, Russia
elvis_75@mail.ru  
alex_vms@gawab.com
*/

#include "tif_config.h"

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#include <errno.h>

#include <stdarg.h>
#include <stdlib.h>
#include <sys/stat.h>

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#ifdef HAVE_FCNTL_H
# include <fcntl.h>
#endif

#ifdef HAVE_IO_H
# include <io.h>
#endif

#include "tiffiop.h"

static tmsize_t
_tiffReadProc(thandle_t fd, void* buf, tmsize_t size)
{
	size_t size_io = (size_t) size;
	if ((tmsize_t) size_io != size)
	{
		errno=EINVAL;
		return (tmsize_t) -1;
	}
	return ((tmsize_t) read((int) fd, buf, size_io));
}

static tmsize_t
_tiffWriteProc(thandle_t fd, void* buf, tmsize_t size)
{
	size_t size_io = (size_t) size;
	if ((tmsize_t) size_io != size)
	{
		errno=EINVAL;
		return (tmsize_t) -1;
	}
	return ((tmsize_t) write((int) fd, buf, size_io));
}

static uint64
_tiffSeekProc(thandle_t fd, uint64 off, int whence)
{
	off_t off_io = (off_t) off;
	if ((uint64) off_io != off)
	{
		errno=EINVAL;
		return (uint64) -1; /* this is really gross */
	}
	return((uint64)lseek((int)fd,off_io,whence));
}

static int
_tiffCloseProc(thandle_t fd)
{
	return(close((int)fd));
}

static uint64
_tiffSizeProc(thandle_t fd)
{
	struct stat sb;
	if (fstat((int)fd,&sb)<0)
		return(0);
	else
		return((uint64)sb.st_size);
}

#ifdef HAVE_MMAP
#include <sys/mman.h>

static int
_tiffMapProc(thandle_t fd, void** pbase, toff_t* psize)
{
	uint64 size64 = _tiffSizeProc(fd);
	tmsize_t sizem = (tmsize_t)size64;
#ifdef VMS
	fsync((int) fd);
#endif
	if ((uint64)sizem==size64) {
		*pbase = (void*)
		    mmap(0, (size_t)sizem, PROT_READ, MAP_SHARED, (int) fd, 0);
		if (*pbase != (void*) -1) {
			*psize = (tmsize_t)sizem;
			return (1);
		}
	}
	return (0);
}

static void
_tiffUnmapProc(thandle_t fd, void* base, toff_t size)
{
	(void) fd;
	(void) munmap(base, (off_t) size);
}
#else /* !HAVE_MMAP */
static int
_tiffMapProc(thandle_t fd, void** pbase, toff_t* psize)
{
	(void) fd; (void) pbase; (void) psize;
	return (0);
}

static void
_tiffUnmapProc(thandle_t fd, void* base, toff_t size)
{
	(void) fd; (void) base; (void) size;
}
#endif /* !HAVE_MMAP */

/*
 * Open a TIFF file descriptor for read/writing.
 */
TIFF*
TIFFFdOpen(int fd, const char* name, const char* mode)
{
	TIFF* tif;

	tif = TIFFClientOpen(name, mode,
	    (thandle_t) fd,
	    _tiffReadProc, _tiffWriteProc,
	    _tiffSeekProc, _tiffCloseProc, _tiffSizeProc,
	    _tiffMapProc, _tiffUnmapProc);
	if (tif)
		tif->tif_fd = fd;
	return (tif);
}

/*
 * Open a TIFF file for read/writing.
 */
TIFF*
TIFFOpen(const char* name, const char* mode)
{
	static const char module[] = "TIFFOpen";
	int m, fd;
	TIFF* tif;

	m = _TIFFgetMode(mode, module);
	if (m == -1)
		return ((TIFF*)0);

/* for cygwin and mingw */
#ifdef O_BINARY
	m |= O_BINARY;
#endif

	fd = open(name, m, 0666, "ctx=bin" /*VMS*/);
	if (fd < 0) {
		TIFFErrorExt(0, module, "%s: Cannot open", name);
		return ((TIFF *)0);
	}

	tif = TIFFFdOpen((int)fd, name, mode);
	if(!tif)
		close(fd);
	return tif;
}


void*
_TIFFmalloc(tmsize_t s)
{
	return (malloc((size_t) s));
}

void
_TIFFfree(void* p)
{
	free(p);
}

void*
_TIFFrealloc(void* p, tmsize_t s)
{
	return (realloc(p, (size_t) s));
}

void
_TIFFmemset(void* p, int v, tmsize_t c)
{
	memset(p, v, (size_t) c);
}

void
_TIFFmemcpy(void* d, const void* s, tmsize_t c)
{
	memcpy(d, s, (size_t) c);
}

int
_TIFFmemcmp(const void* p1, const void* p2, tmsize_t c)
{
	return (memcmp(p1, p2, (size_t) c));
}

static void
vmsWarningHandler(const char* module, const char* fmt, va_list ap)
{
	if (module != NULL)
		fprintf(stderr, "%s: ", module);
	fprintf(stderr, "Warning, ");
	vfprintf(stderr, fmt, ap);
	fprintf(stderr, ".\n");
}
TIFFErrorHandler _TIFFwarningHandler = vmsWarningHandler;

static void
vmsErrorHandler(const char* module, const char* fmt, va_list ap)
{
	if (module != NULL)
		fprintf(stderr, "%s: ", module);
	vfprintf(stderr, fmt, ap);
	fprintf(stderr, ".\n");
}
TIFFErrorHandler _TIFFerrorHandler = vmsErrorHandler;

/* vim: set ts=8 sts=8 sw=8 noet: */

#if !HAVE_IEEEFP
/* IEEE floting point handling */

typedef struct ieeedouble {
    unsigned long       mant2;                  /* fix NDR: full 8-byte swap */
    unsigned long       mant    : 20,
                exp     : 11,
                sign    : 1;
} ieeedouble;
typedef struct ieeefloat {
        unsigned long
                mant    : 23,
                exp     : 8,
                sign    : 1;
} ieeefloat;

/*
 * NB: These are D_FLOAT's, not G_FLOAT's. A G_FLOAT is
 *  simply a reverse-IEEE float/double.
 */

typedef struct {
 unsigned long  mant1   : 7,
                exp     : 8,
                sign    : 1,
                mant2   : 16,
                mant3   : 16,
                mant4   : 16;
} nativedouble;
typedef struct {
  unsigned long mant1   : 7,
                exp     : 8,
                sign    : 1,
                mant2   : 16;
} nativefloat;

typedef union {
        ieeedouble      ieee;
        nativedouble    native;
        char            b[8];
        uint32          l[2];
        double          d;
} double_t;

typedef union {
        ieeefloat       ieee;
        nativefloat     native;
        char            b[4];
        uint32          l;
        float           f;
} float_t;

#if defined(VAXC) || defined(DECC)
#pragma inline(ieeetod,dtoieee)
#endif

/*
 * Convert an IEEE double precision number to native double precision.
 * The source is contained in two longwords, the second holding the sign,
 * exponent and the higher order bits of the mantissa, and the first
 * holding the rest of the mantissa as follows:
 * (Note: It is assumed that the number has been eight-byte swapped to
 * LSB first.)
 *
 * First longword:
 *      32 least significant bits of mantissa
 * Second longword:
 *      0-19:   20 most significant bits of mantissa
 *      20-30:  exponent
 *      31:     sign
 * The exponent is stored as excess 1023.
 * The most significant bit of the mantissa is implied 1, and not stored.
 * If the exponent and mantissa are zero, the number is zero.
 * If the exponent is 0 (i.e. -1023) and the mantissa is non-zero, it is an
 * unnormalized number with the most significant bit NOT implied.
 * If the exponent is 2047, the number is invalid, in case the mantissa is zero,
 * this means overflow (+/- depending of the sign bit), otherwise
 * it simply means invalid number.
 *
 * If the number is too large for the machine or was specified as overflow,
 * +/-HUGE_VAL is returned.
 */
INLINE static void
ieeetod(double *dp)
{
        double_t source;
        long sign,exp,mant;
        double dmant;

        source.ieee = ((double_t*)dp)->ieee;
        sign = source.ieee.sign;
        exp = source.ieee.exp;
        mant = source.ieee.mant;

        if (exp == 2047) {
                if (mant)                       /* Not a Number (NAN) */
                        *dp = HUGE_VAL;
                else                            /* +/- infinity */
                        *dp = (sign ? -HUGE_VAL : HUGE_VAL);
                return;
        }
        if (!exp) {
                if (!(mant || source.ieee.mant2)) {     /* zero */
                        *dp=0;
                        return;
                } else {                        /* Unnormalized number */
                        /* NB: not -1023, the 1 bit is not implied */
                        exp= -1022;
                }
        } else {
                mant |= 1<<20;
                exp -= 1023;
        }
        dmant = (((double) mant) +
                ((double) source.ieee.mant2) / (((double) (1<<16)) *
                ((double) (1<<16)))) / (double) (1<<20);
        dmant = ldexp(dmant, exp);
        if (sign)
                dmant= -dmant;
        *dp = dmant;
}

INLINE static void
dtoieee(double *dp)
{
        double_t num;
        double x;
        int exp;

        num.d = *dp;
        if (!num.d) {                   /* Zero is just binary all zeros */
                num.l[0] = num.l[1] = 0;
                return;
        }

        if (num.d < 0) {                /* Sign is encoded separately */
                num.d = -num.d;
                num.ieee.sign = 0;
        }

        /* Now separate the absolute value into mantissa and exponent */
        x = frexp(num.d, &exp);

        /*
         * Handle cases where the value is outside the
         * range for IEEE floating point numbers.
         * (Overflow cannot happen on a VAX, but underflow
         * can happen for G float.)
         */
        if (exp < -1022) {              /* Unnormalized number */
                x = ldexp(x, -1023-exp);
                exp = 0;
        } else if (exp > 1023) {        /* +/- infinity */
                x = 0;
                exp = 2047;
        } else {                        /* Get rid of most significant bit */
                x *= 2;
                x -= 1;
        }
        num.ieee.exp = exp;

        x *= (double) (1<<20);
        num.ieee.mant = (long) x;
        x -= (double) num.ieee.mant;
        num.ieee.mant2 = (long) (x*((double) (1<<16)*(double) (1<<16)));

        if (!(num.ieee.mant || num.ieee.exp || num.ieee.mant2)) {
                /* Avoid negative zero */
                num.ieee.sign = 0;
        }
        ((double_t*)dp)->ieee = num.ieee;
}

/*
 * Beware, these do not handle over/under-flow
 * during conversion from ieee to native format.
 */
#define NATIVE2IEEEFLOAT(fp) { \
    float_t t; \
    if (t.ieee.exp = (fp)->native.exp) \
        t.ieee.exp += -129 + 127; \
    t.ieee.sign = (fp)->native.sign; \
    t.ieee.mant = ((fp)->native.mant1<<16)|(fp)->native.mant2; \
    *(fp) = t; \
}
#define IEEEFLOAT2NATIVE(fp) { \
    float_t t; int v = (fp)->ieee.exp; \
    if (v) v += -127 + 129;             /* alter bias of exponent */\
    t.native.exp = v;                   /* implicit truncation of exponent */\
    t.native.sign = (fp)->ieee.sign; \
    v = (fp)->ieee.mant; \
    t.native.mant1 = v >> 16; \
    t.native.mant2 = v;\
    *(fp) = t; \
}

#define IEEEDOUBLE2NATIVE(dp) ieeetod(dp)

#define NATIVE2IEEEDOUBLE(dp) dtoieee(dp)


/*
 * These unions are used during floating point
 * conversions.  The above macros define the
 * conversion operations.
 */
void
TIFFCvtIEEEFloatToNative(TIFF* tif, u_int n, float* f)
{
        float_t* fp = (float_t*) f;

        while (n-- > 0) {
                IEEEFLOAT2NATIVE(fp);
                fp++;
        }
}

void
TIFFCvtNativeToIEEEFloat(TIFF* tif, u_int n, float* f)
{
        float_t* fp = (float_t*) f;

        while (n-- > 0) {
                NATIVE2IEEEFLOAT(fp);
                fp++;
        }
}
void
TIFFCvtIEEEDoubleToNative(TIFF* tif, u_int n, double* f)
{
        double_t* fp = (double_t*) f;

        while (n-- > 0) {
                IEEEDOUBLE2NATIVE(fp);
                fp++;
        }
}

void
TIFFCvtNativeToIEEEDouble(TIFF* tif, u_int n, double* f)
{
        double_t* fp = (double_t*) f;

        while (n-- > 0) {
                NATIVE2IEEEDOUBLE(fp);
                fp++;
        }
}
#endif




